@@ -17,16 +17,23 @@ from boto.exception import DynamoDBResponseError
 
 
 log = logging.getLogger(__name__)
+dblog = logging.getLogger(__name__+".database-access")
 
 
 MAX_RETRIES = 100
-
+# primary key of the magic item used for autoinc
+MAGIC_KEY = -1
 
 class SchemaError(Exception):
     """Raised when a DynamoDBModel class's schema is incorrect."""
     pass
 
 
+class ThroughputError(Exception):
+    """Raised when reaquested throughput can not be allocated."""
+    pass
+
+
 class MaxRetriesExceededError(Exception):
     """Raised when a failed operation couldn't be completed after retrying
     MAX_RETRIES times (e.g. saving an autoincrementing hash_key).
@@ -64,13 +71,16 @@ class autoincrement_int(int):
 
     Auto-incrementing int keys are implemented by storing a special "magic"
     item in the table with the following properties:
-      - hash_key_value = 0
+      - hash_key_value = -1
       - __max_hash_key__ = X
     where X is the maximum used hash_key value.
 
-    Inserting a new item issues a conditional write to the magic item,
-    incrementing __max_hash_key__ iff it hasn't changed and using that new
-    value for the created object's hash_key value.
+    Inserting a new item issues an atomic add on the '__max_hash_key__' value.
+    Its new value is returned and used as the primary key for the new elem.
+
+    Note that hash_key_value is set to '-1' while '__max_hash_key__' initial
+    value is 0. This will element at key '0' unused. It's actually a garbage item
+    for cases where a value is manually added to an unitialized index.
     """
     pass
 
@@ -112,17 +122,36 @@ def _get_proto_value(schema_type):
     return schema_type()
 
 
-def _get_default_value(schema_type):
-    """Return a default value matching schema_type:
+def _get_default_value(schema_type, default=None):
+    """Return a default value matching schema_type or default if provided:
 
-      - For datetime.datetime, it's 1970-01-01 00:00:00 UTC.
+      - For datetime.datetime, it's NOW.
       - For container types, it's an empty container.
       - For strings, it's an empty string.
       - For numbers, it's zero.
+
+      This function may raise TypeError exception if:
+       - default was callable and required arguments
+       - default or its return value is not an instance of schema_type
+
+      :param schema_type class object to instanciate
+      :param default default value. May be a value or a callable (functions, class, ...) It must *NOT* require an any argument and it's type must match schema_type
+
     """
+    if default is not None:
+        # If default is callable(function, constructor, ...), try to dereference it
+        if hasattr(default, '__call__'):
+            # Might raise a "TypeError" if arguments were needed
+            default = default()
+        # Check default value consitency
+        if not isinstance(default, schema_type):
+            raise TypeError("Expected default value of type {}, got: {}".format(schema_type, type(default)))
+        else:
+            return default
+
     if schema_type is datetime:
-        # Beginning of the UNIX epoch
-        return datetime(1970, 1, 1, 0, 0, 0, tzinfo=utc_tz)
+        # Current Timestamp
+        return datetime.now(tz=utc_tz)
 
     return schema_type()
 
@@ -155,6 +184,12 @@ def _python_to_dynamodb(value):
         s = s[:-2] + ':' + s[-2:]
         return s
 
+    # This case prevents `'fields': False` to be added when genereating expected
+    # values dict in save as this would mean 'field does not exist' instead of
+    # 'field exists and is False'.
+    if isinstance(value, bool):
+        return int(value)
+
     if value or value == 0:
         return value
 
@@ -184,12 +219,7 @@ def _dynamodb_to_python(schema_type, value):
     # This means a missing attribute for a _JSON_TYPES object is valid,
     # and results in an empty sequence. Is that a bad thing?
     if value is None:
-        if schema_type is datetime:
-            # Default value for unsupplied datetimes:
-            # the beginning of the UNIX epoch.
-            return datetime.now(tz=utc_tz)
-        # Empty sets/strings are represented as missing attributes
-        return schema_type()
+        return None
 
     if schema_type in _JSON_TYPES:
         return schema_type(json.loads(value))
@@ -211,7 +241,10 @@ def _dynamodb_to_python(schema_type, value):
 class ConnectionBorg(object):
     """Borg that handles access to DynamoDB.
 
-    You should never make any explicit/direct boto.dynamodb calls by yourself.
+    You should never make any explicit/direct boto.dynamodb calls by yourself
+    except for table maintenance operations :
+    * boto.dynamodb.table.update_throughput()
+    * boto.dynamodb.table.delete()
 
     Remember to call :meth:`set_auth_credentials`, or to set the
     ``AWS_ACCESS_KEY_ID`` and ``AWS_SECRET_ACCESS_KEY`` environment variables
@@ -244,7 +277,7 @@ class ConnectionBorg(object):
             return self._connections[thread_id]
 
     def _create_autoincrement_magic_item(self, table):
-        item = table.new_item(hash_key=0, attrs={
+        item = table.new_item(hash_key=MAGIC_KEY, attrs={
             "__max_hash_key__": 0
         })
         # Conditional write: don't risk overwriting the DB.
@@ -287,6 +320,9 @@ class ConnectionBorg(object):
         if not cls.__schema__:
             raise SchemaError("Class does not define __schema__", cls)
 
+        if read_units < 5 or write_units < 5:
+            raise ThroughputError("Either read or write throughput is below 5")
+
         hash_key_type = cls.__schema__[hash_key_name]
 
         if hash_key_type is autoincrement_int:
@@ -349,6 +385,13 @@ class DynamoDBModel(object):
           Supported attribute_types are: int, long, float, str, unicode, set.
           Default values are obtained by calling the type with no args
           (so 0 for numbers, "" for strings and empty sets).
+      - __defaults__: {attribute_name: defaulter} optional mapping.
+          This dict allows to provide a default value for each attribute_name at
+          object creation time. It will *never* be used when loading from the DB.
+          It is fully optional. If no value is supplied the empty value
+          corresponding to the type will be used.
+          "defaulter" may either be a scalar value or a callable with no
+          arguments.
 
     To redefine serialization/deserialization semantics (e.g. to have more
     complex schemas, like auto-serialized JSON data structures), override the
@@ -377,6 +420,28 @@ class DynamoDBModel(object):
     __hash_key__ = None
     __range_key__ = None
     __schema__ = None
+    __defaults__ = {}
+
+    def __init__(self, **kwargs):
+        """Create an instance of the model. All fields defined in the schema
+        are created. By order of prioritym its value will be loaded from:
+
+        * kwargs
+        * __default__
+        * mapper's default (0, empty string, empty set, ...)
+
+        We're supplying this method to avoid the need for extra checks in save and
+        ease object initial creation.
+        """
+        defaults = type(self).__defaults__
+        schema = type(self).__schema__
+
+        for (name, type_) in schema.iteritems():
+            if name in kwargs:
+                default = kwargs.get(name)
+            else:
+                default = defaults[name] if name in defaults else None
+            setattr(self, name, _get_default_value(type_, default))
 
     @classmethod
     def from_dict(cls, d):
@@ -386,10 +451,13 @@ class DynamoDBModel(object):
         Default values are used for anything that's missing from the dict
         (see DynamoDBModel class docstring).
         """
+        #FIXME: rename to from_db_dict when breaking compat or disambiguing
         instance = cls()
         for (name, type_) in cls.__schema__.iteritems():
-            # Fill in missing attributes with default values.
+            # Set the value if we got one from DynamoDB. Otherwise, stick with the default
             value = _dynamodb_to_python(type_, d.get(name))
+            if value is None:
+                value = _get_default_value(type_)
             setattr(instance, name, value)
         return instance
 
@@ -418,11 +486,14 @@ class DynamoDBModel(object):
         else:
             r_value = None
 
-        return cls.from_dict(
-            table.get_item(
-                hash_key=h_value,
-                range_key=r_value,
-                consistent_read=consistent_read))
+        item = table.get_item(
+                    hash_key=h_value,
+                    range_key=r_value,
+                    consistent_read=consistent_read)
+
+        dblog.debug("[GET] %s(%s, %s)", cls.__table__, h_value, r_value)
+
+        return cls.from_dict(item)
 
     @classmethod
     def get_batch(cls, keys):
@@ -438,7 +509,12 @@ class DynamoDBModel(object):
             (e.g. [("user1", 1), ("user1", 2), ("user1", 3)]).
 
         get_batch *always* performs eventually consistent reads.
+
+        Please not that a batch can *not* read more than 100 items at once.
         """
+        if len(keys) > 100:
+            raise ValueError("Too many items to read in a single batch. Maximum is 100.")
+
         borg = ConnectionBorg()
         table = borg.get_table(cls.__table__)
         # Convert all the keys to DynamoDB values.
@@ -456,12 +532,15 @@ class DynamoDBModel(object):
         batch_list.add_batch(table, dynamo_keys)
 
         res = batch_list.submit()
+
+        dblog.debug("[BATCH GET] %s", cls.__table__)
+
         return [
             cls.from_dict(d) for d in res[u"Responses"][cls.__table__][u"Items"]
         ]
 
     @classmethod
-    def query(cls, hash_key_value, range_key_condition=None, consistent_read=False):
+    def query(cls, hash_key_value, range_key_condition=None, consistent_read=False, reverse=False):
         """Query DynamoDB for items matching the requested key criteria.
 
         You need to supply an exact hash key value, and optionally, conditions
@@ -486,13 +565,16 @@ class DynamoDBModel(object):
         """
         table = ConnectionBorg().get_table(cls.__table__)
         h_value = _python_to_dynamodb(hash_key_value)
-        return (
-            cls.from_dict(d)
-            for d in table.query(
+
+        res = table.query(
                 h_value,
                 range_key_condition,
-                consistent_read=consistent_read)
-        )
+                consistent_read=consistent_read,
+                scan_index_forward=not reverse)
+
+        dblog.debug("[QUERY] %s(%s, %s)", cls.__table__, h_value, range_key_condition)
+
+        return (cls.from_dict(d) for d in res)
 
     @classmethod
     def scan(cls, scan_filter=None):
@@ -510,20 +592,18 @@ class DynamoDBModel(object):
         :rtype: generator
         """
         table = ConnectionBorg().get_table(cls.__table__)
+        hash_key_name = table.schema.hash_key_name
+
+        res = table.scan(scan_filter)
+
+        dblog.debug("[QUERY] %s(%s)", cls.__table__, scan_filter)
+
         return (
             cls.from_dict(d)
-            for d in table.scan(scan_filter)
+            for d in res
+            if d['key'] != MAGIC_KEY or cls.__schema__[hash_key_name] != autoincrement_int
         )
 
-    def __init__(self):
-        """Create a default empty instance of the class with default values
-        according to its schema.
-
-        We're supplying this method to avoid the need for extra checks in save.
-        """
-        for (name, type_) in type(self).__schema__.iteritems():
-            setattr(self, name, _get_default_value(type_))
-
     def to_db_dict(self):
         """Return a dict representation of the object according to the class's
         schema, suitable for direct storage in DynamoDB.
@@ -564,33 +644,38 @@ class DynamoDBModel(object):
     def _save_autoincrement_hash_key(self, item):
         """Compute an autoincremented hash_key for an item and save it to the DB.
 
-        TODO Add schema checks.
+        To achieve this goal, we keep a special object at hash_key=0 to keep track
+        of the counter status. We then issue an atomic inc to the counter field.
+        We do not need to read it befor as we know its hesh_key yet.
+        The new value is send back to us and used as the hash_key for elem
         """
+        counter_key = '__max_hash_key__'
         tries = 0
+
         while tries < MAX_RETRIES:
             tries += 1
-            max_hash_item = item.table.get_item(0, consistent_read=True)
-            max_hash_key = max_hash_item["__max_hash_key__"]
-            max_hash_item["__max_hash_key__"] += 1
+            # Create a 'new item' with key=0 and trigger an atomic increment
+            # This spares one read unit :)
+            max_hash_item = item.table.new_item(MAGIC_KEY)
+            max_hash_item.add_attribute(counter_key, 1)
+            max_hash_item = max_hash_item.save(return_values='ALL_NEW')
+            # We just reserved that value for the hash key
+            item[item.hash_key_name] = max_hash_item['Attributes'][counter_key]
+
             try:
-                # Conditional write: we're overwriting iff the value
-                # hasn't changed.
-                max_hash_item.put({"__max_hash_key__": max_hash_key})
-                # We just reserved that value for the hash key
-                item[item.hash_key_name] = max_hash_key + 1
-                # If despite our checks someone manually inserted an item at
-                # that primary key, we'll retry with the next available value.
+                # Make sure this primary key was not 'stolen' by a direct DB access
                 item.put({item.hash_key_name: False})
-                break
+                return
             except DynamoDBResponseError as e:
                 if e.error_code != "ConditionalCheckFailedException":
                     # Unhandled exception
                     raise
                 log.debug(
-                    "table=%s, max hash key has changed. Retrying (%s).",
-                    item.table, tries)
-        else:
-            raise MaxRetriesExceededError()
+                    "table=%s, An item seems to have been manually inserted at index %s (%s).",
+                    item.table, item[item.hash_key_name], tries)
+
+        # This table auto-incr has been screwed up...
+        raise MaxRetriesExceededError()
 
     def save(self, allow_overwrite=True, expected_values=None):
         """Save the object to the database.
@@ -619,11 +704,16 @@ class DynamoDBModel(object):
         item_data = self.to_db_dict()
         item = Item(table, attrs=item_data)
 
-        if schema[hash_key] is autoincrement_int and item_data[hash_key] == 0:
-            # We're inserting a new item in an autoincrementing table.
+        # Detect magic elem manual overwrite
+        if schema[hash_key] == autoincrement_int and item_data[hash_key] == MAGIC_KEY:
+            raise SchemaError()
+        # We're inserting a new item in an autoincrementing table.
+        if schema[hash_key] == autoincrement_int and item_data[hash_key] == 0:
+            # Compute the index and save the object
             self._save_autoincrement_hash_key(item)
             # Update the primary key so that it reflects what it was saved as.
             setattr(self, hash_key, item[hash_key])
+        # Regular save
         else:
             if allow_overwrite:
                 expected_db_values = {}
